// Standard Includes for MPI, C and OS calls
#include <mpi.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

// C++ standard I/O calls
#include <iostream>
using std::cerr ;
using std::cout ;
using std::endl ;

#include <cmath>
using std::pow ;

int sleep_time = 540 ; // 9 minutes

//**************************************************************************/
// UTILITY ROUTINES


// This routine handles program traps.  We put this in here so that the
// mpi program will be more robust to crashes and won't lock up the queue

void program_trap(int sig) /*,code,scp,addr)*/ {
  const char *sigtype = "(undefined)" ;
  
  switch(sig) {
  case SIGBUS:
    sigtype = "a Bus Error" ;
    break ;
  case SIGSEGV:
    sigtype = "a Segmentation Violation" ;
    break ;
  case SIGILL:
    sigtype = "an Illegal Instruction Call" ;
    break ;
  case SIGSYS:
    sigtype = "an Illegal System Call" ;
    break ;
  case SIGFPE:
    sigtype = "a Floating Point Exception" ;
    break ;
  case SIGALRM:
    sigtype = "a Alarm Signal!" ;
    break ;
  }
  fprintf(stderr,"ERROR: Program terminated due to %s\n",sigtype) ;
  abort() ;
}

// This routine redirects the OS signal traps to the above routine
// Also sets a 9 minute alarm to kill the program in case it runs
// away.  This really should prevent job queue problems for errant
// programs.
void chopsigs_() {
  signal(SIGBUS,program_trap) ;
  signal(SIGSEGV,program_trap) ;
  signal(SIGILL,program_trap) ;
  signal(SIGSYS,program_trap) ;
  signal(SIGFPE,program_trap) ;
  signal(SIGALRM,program_trap) ;
  // Send an alarm signal after 9 minutes elapses
  alarm(sleep_time) ;
}


// A utility routine for measuring time
double get_timer() {
  static double to = 0;
  double tn,t ;
  tn = MPI_Wtime() ;
  t = tn - to ;
  to = tn ;
  return t ;
}

// function that implements 2^i
inline int pow2(int p) {
  return 1<<p ;
}

// function that implements log_2(i)
inline int log2(int v) {
  int d = 0 ;
  for(v=v>>1;v!=0;v=v>>1)
    d++ ;
  return d ;
}

//************************************************************************/
// Communication Variables

// We will use these routine to store the size and rank in MPI_COMM_WORLD
int numprocs,myid ;

//***********************************************************************/
// Example implementation of bitonic sort

// Below are the compare split functions for use in the bitonic sort.
// They take in three buffers, a local buffer, and two temporary buffers
// They use pointer swapping to avoid copying.
void compare_split_max(double **local_buf, double **recv_buf,
                       double **result_buf,
                       int loc_size, int max_buf_size,
                       int dest_addr) {
  MPI_Status stat ;

  MPI_Sendrecv(*local_buf,loc_size,MPI_DOUBLE,dest_addr,0,
               *recv_buf,max_buf_size,MPI_DOUBLE,dest_addr,0,
               MPI_COMM_WORLD,&stat) ;

  int recv_size ;
  MPI_Get_count(&stat,MPI_DOUBLE,&recv_size) ;

  int ld=loc_size-1,rd=recv_size-1 ;
  for(int i=loc_size-1;i>=0;--i) {
    if(rd<0)
      (*result_buf)[i] = (*local_buf)[ld--] ;
    else if(ld<0)
      (*result_buf)[i] = (*recv_buf)[rd--] ;
    else if((*local_buf)[ld] > (*recv_buf)[rd])
      (*result_buf)[i] = (*local_buf)[ld--] ;
    else
      (*result_buf)[i] = (*recv_buf)[rd--] ;
  }

  std::swap(*result_buf,*local_buf) ;

}

void compare_split_min(double **local_buf, double **recv_buf,
                       double **result_buf,
                       int loc_size, int max_buf_size,
                       int dest_addr) {
  MPI_Status stat ;
  MPI_Sendrecv(*local_buf,loc_size,MPI_DOUBLE,dest_addr,0,
               *recv_buf,max_buf_size,MPI_DOUBLE,dest_addr,0,
               MPI_COMM_WORLD,&stat) ;

  int recv_size ;
  MPI_Get_count(&stat,MPI_DOUBLE,&recv_size) ;

  int ld=0,rd=0 ;
  for(int i=0;i<loc_size;++i) {
    if(rd==recv_size)
      (*result_buf)[i] = (*local_buf)[ld++] ;
    else if(ld==loc_size)
      (*result_buf)[i] = (*recv_buf)[rd++] ;
    else if((*local_buf)[ld] < (*recv_buf)[rd]) 
      (*result_buf)[i] = (*local_buf)[ld++] ;
    else
      (*result_buf)[i] = (*recv_buf)[rd++] ;
  }
  std::swap(*result_buf,*local_buf) ;
}

// Parallel bitonic sort algorithm as described in the text.
double *parallel_bitonic_sort(double *buffer, int &loc_buf_size,
                              int max_size) {
  if(numprocs & (numprocs-1)) {
    cerr << "bitonic sort requires 2^d processors" << endl ;
    MPI_Abort(MPI_COMM_WORLD,-1) ;
    abort() ;
  }  
  
  // First sort the local sequence
  std::sort(buffer, buffer+loc_buf_size) ;

  // Allocate some buffers ;
  double *recv_buffer = new double[max_size] ;
  double *result_buffer = new double[max_size] ;
  
  // Perform bitonic sort on a d dimensional hypercube
  int d = log2(numprocs) ;

  for(int i=0;i<d;++i) {
    for(int j=i;j>=0;--j) {
      int ibit = (myid & pow2(i+1)) != 0 ;
      int jbit = (myid & pow2(j)) != 0 ;
      int dest_addr = myid ^ pow2(j) ;
      if(ibit != jbit)
        compare_split_max(&buffer,&recv_buffer,&result_buffer,
                          loc_buf_size,max_size, dest_addr);
      else
        compare_split_min(&buffer,&recv_buffer,&result_buffer,
                          loc_buf_size,max_size, dest_addr);
    }
  }

  delete[] recv_buffer ;
  delete[] result_buffer ;
  return buffer ;
}

//***************************************************************************/
// Program Testing

// Check to see if the sequence has been sorted.  We do this by counting the
// number of times that val[i]>val[i+1].  This will be zero when the sequence
// is properly sorted.
void check_sort(double local_numbers[], int local_size) {
  int check_local=0 ;
  for(int i=0;i<local_size-1;++i)
    if(local_numbers[i]>local_numbers[i+1])
      check_local++ ;
  if(myid != 0) { 
    MPI_Status stat ;
    double lowval ;
    MPI_Recv(&lowval,1,MPI_DOUBLE,myid-1,0,MPI_COMM_WORLD,&stat) ;
    if(lowval>local_numbers[0]) {
      check_local++ ;
    }
  }
  if(myid+1 != numprocs) { //last processor doesn't have anyone to send to
    MPI_Send(&local_numbers[local_size-1],1,MPI_DOUBLE,myid+1,0,MPI_COMM_WORLD) ;
  }

  int check ;
  MPI_Reduce(&check_local,&check,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD) ;
  if(myid == 0) {
    cout << check << " errors in sorting" << endl ;
  }
}



//****************************************************************************/
// Main code starts here.

main(int argc, char **argv) {
  double time_passes, max_time ;
  /* Initialize MPI */
  MPI_Init(&argc,&argv) ;

  // Handle signals (Segmentation faults, etc.  Make sure to call MPI cleanup
  // routines when things go wrong.
  chopsigs_() ;
  
  /* Get the number of processors and my processor identification */
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs) ;
  MPI_Comm_rank(MPI_COMM_WORLD,&myid) ;

  int input_size = 1024 ; // For debugging we will sort 1024 numbers
  sleep_time=120 ;
  
  if(argc == 2) {
    input_size = atoi(argv[1]) ;
    sleep_time = 540 ;
  }

  // Output diagnostic indicating start of operations
  if(0 == myid) {
    cout << "Starting " << numprocs << " processors." << endl ;
    cout << "generating input sequence consisting of " <<input_size
         << " doubles."<< endl ;
  }

  // Compute the number of numbers that will be allocated to this processor
  // This estimates the size of the initial buffer sizes.  (Note, this doesn't
  // estimate final buffer sizes!
  int local_input_size = input_size/numprocs ;
  int max_local_size = local_input_size+1 ;
  
  // spread the remainder evenly over processors.
  int remainder = input_size%numprocs ;
  if(myid < remainder)  
    local_input_size += 1 ;

  // Allocate space for the local numbers
  double *local_numbers = new double[max_local_size] ;

  // Call MPI_Barrier to before measuring time.  It doesn't assure time sync
  // but its better than nothing.
  MPI_Barrier(MPI_COMM_WORLD) ;

  /* Every call to get_timer resets the stopwatch.  The next call to get_timer
     will return the amount of time since now */
  get_timer() ;

  // Now we are about to generate a random sequence of numbers.  To make
  // performance measurements fair, we will generate the numbers in such
  // a way that the sequence will always be the same regardless of the number
  // of processors.  We do this by passing the random number seed between
  // processors.  This approach works but isn't parallel.  Parallel random
  // number generation is an interesting problem, but not the one this project
  // is investigating.
  
  // First we recieve the seed to begin computing our random numbers
  // locally.

  // The root processor generates the initial seed
  static unsigned short xi[4] = {0,0,1,0} ;

  // Get the seed to what the previous processor had computed if I am not the
  // root processor.
  if(myid != 0) { 
    MPI_Status stat ;
    MPI_Recv(xi,4,MPI_UNSIGNED_SHORT,myid-1,0,MPI_COMM_WORLD,&stat) ;
  }

  // Distribution type, comment out ODD_DIST to get a uniform distribution
  // Undergradutes comment this line, graduates leave it in.
#define ODD_DIST

  for(int i=0;i<local_input_size;++i) {
    xi[3] += 1 ;
    double val = erand48(xi) ;
#ifdef ODD_DIST
    double p = double(xi[3])/double(input_size) ;
    val = pow(val,1.0+3*p) ;
    val = val*val ;
#endif
    local_numbers[i] = val ;
  }

  // Send current seed to next processors
  if(myid+1 != numprocs) { //last processor doesn't have anyone to send to
    MPI_Send(xi,4,MPI_UNSIGNED_SHORT,myid+1,0,MPI_COMM_WORLD) ;
  }

  // Synchronize for a time measurement.
  MPI_Barrier(MPI_COMM_WORLD) ;

  /* Every call to get_timer resets the stopwatch.  The next call to get_timer
     will return the amount of time since now */
  double seq_gen_time = get_timer() ;
  
  MPI_Reduce(&seq_gen_time, &max_time, 1, MPI_DOUBLE,MPI_MAX, 0,
             MPI_COMM_WORLD) ;

  if(myid == 0) {
    cout << "completed generation of a sequence of size " << input_size 
         << "." << endl ;
    cout << "sequence generation required " << max_time
         << " seconds." << endl ;
  }

  MPI_Barrier(MPI_COMM_WORLD) ;
  get_timer() ;

  // At this point every processor has a local list of random double precision
  // numbers distributed evenly over the interval (0,1).   The local list of
  // numbers is stored in the array local_numbers, where there are
  // local_input_size numbers in the array.  It is your task to sort these
  // numbers using the enumeration provided by MPI_Rank().
  
  //**************************************************************************/
  // Insert your parallel sort here, I've included the bitonic sort
  // for a simple example.
  //**************************************************************************/

  local_numbers = parallel_bitonic_sort(local_numbers,local_input_size,
                                        max_local_size) ;

  // Find out how much time it took to sort.
  double par_sort_time = get_timer() ;
  MPI_Reduce(&par_sort_time, &max_time, 1, MPI_DOUBLE,MPI_MAX, 0,
             MPI_COMM_WORLD) ;
  if(myid == 0) {
    cout << "parallel sort time = " << max_time << endl ;
  }


  // Check to see if sequence is actually sorted.
  check_sort(local_numbers,local_input_size) ;

  // Clean things up and exit.
  MPI_Finalize() ;
  return 0 ;
  
}
